//    This file is part of JustMotion.

//    JustMotion is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    JustMotion is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

#include "camera.h"

Camera::Camera(const QString &confFile, const QString &statDir, ProcControl *proc, QCoreApplication *parent) : QObject(nullptr)
{
    Q_UNUSED(parent);

    fsW       = new QFileSystemWatcher(this);
    statTimer = new QTimer(this);
    statPath  = statDir;
    evtLoop   = nullptr;
    detLoop   = nullptr;
    recLoop   = nullptr;
    delOnZero = false;
    objCount  = 0;

    statTimer->setInterval(5000);
    statTimer->start();
    proc->objPlusOne();

    connect(fsW,       &QFileSystemWatcher::fileChanged, this, &Camera::confChanged);
    connect(proc,      &ProcControl::prepForClose,       this, &Camera::prepForDel);
    connect(this,      &Camera::destroyed,               proc, &ProcControl::objMinusOne);
    connect(statTimer, &QTimer::timeout,                 this, &Camera::updateStat);

    start(confFile);
}

void Camera::objMinusOne()
{
    objCount--;

    if (objCount == 0)
    {
        thr1->deleteLater();
        thr2->deleteLater();
        thr3->deleteLater();

        if (delOnZero)
        {
            QDir(shared.buffPath).removeRecursively();

            deleteLater();
        }
    }
}

void Camera::prepForDel()
{
    statTimer->blockSignals(true);

    delOnZero = true;

    emit stop();
}

void Camera::updateStat()
{
    auto statFile = statPath + "/" + shared.camName;

    QFile file(statFile);

    file.open(QFile::WriteOnly);
    file.write(statusLine().toUtf8());
    file.close();
}

int Camera::start(const QString &conf)
{
    if (rdConf(conf, &shared))
    {
        setupBuffDir(shared.buffPath, true);

        if (!fsW->files().contains(conf))
        {
            fsW->addPath(conf);
        }

        thr1 = new QThread(nullptr);
        thr2 = new QThread(nullptr);
        thr3 = new QThread(nullptr);

        evtLoop = new EventLoop(&shared,  thr1, nullptr);
        detLoop = new DetectLoop(&shared, thr2, nullptr);
        recLoop = new RecordLoop(&shared, thr3, nullptr);

        connect(this, &Camera::stop, thr1, &QThread::quit);
        connect(this, &Camera::stop, thr2, &QThread::quit);
        connect(this, &Camera::stop, thr3, &QThread::quit);

        connect(thr1, &QThread::finished, this, &Camera::objMinusOne);
        connect(thr2, &QThread::finished, this, &Camera::objMinusOne);
        connect(thr3, &QThread::finished, this, &Camera::objMinusOne);

        connect(detLoop, &DetectLoop::starving, recLoop, &RecordLoop::restart);

        thr1->start();
        thr2->start();
        thr3->start();

        objCount = 3;
    }

    return shared.retCode;
}

void Camera::confChanged(const QString &path)
{
    emit stop();

    if (!QFileInfo::exists(path))
    {
        deleteLater();
    }
    else
    {
        auto ret = start(path);

        if (ret != 0)
        {
            deleteLater();
        }
    }
}

QString Camera::statusLine()
{
    return "Detection: "       + detLoop->statusLine() +
           " Event-Scraping: " + evtLoop->statusLine() +
           " Recording: "      + recLoop->statusLine() +
           " Name: "           + shared.camName + "\n";
}
